package org.apidesign.apifest08.test;

import java.math.BigDecimal;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.ConvertorFactory;
import org.apidesign.apifest08.currency.ExchangeRate;
import org.apidesign.apifest08.currency.InvalidConversionException;
import org.apidesign.apifest08.currency.TimedConvertor;

/** The exchange rates are not always the same. They are changing. However
 * as in order to predict the future, one needs to understand own past. That is
 * why it is important to know the exchange rate as it was at any time during
 * the past.
 * <p>
 * Today's quest is to enhance the convertor API to deal with dates.
 * One shall be able to convert a currency at any date. Each currencies rate shall
 * be associated with a range between two Date objects. In order
 * to keep compatibility with old API that knew nothing about dates, the
 * rates associated then are applicable "for eternity". Any use of existing
 * convert methods that do not accept a Date argument, uses the current
 * System.currentTimeMillis() as default date.
 */
public class Task4Test extends TestCase {
    public Task4Test(String testName) {
        super(testName);
    }

    private Calendar gmtCalendar;
    
    @Override
    protected void setUp() throws Exception {
        gmtCalendar = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    }

    @Override
    protected void tearDown() throws Exception {
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //

    /** Takes a convertor with any rates associated and creates new convertor
     * that returns the same values as the old one for time between from to till.
     * Otherwise it returns no results. This is just a helper method that
     * shall call some real one in the API.
     * 
     * @param old existing convertor
     * @param from initial date (inclusive)
     * @param till final date (exclusive)
     * @return new convertor
     */
    public static Convertor limitTo(Convertor old, Date from, Date till) {
        final Convertor convertor;

        convertor = ConvertorFactory.getConvertor(from, till, old);
        
        return convertor;
    }


    public void testCompositionOfLimitedConvertors() throws Exception {

        gmtCalendar.set(2008, Calendar.OCTOBER, 1, 0, 0, 0);
        Date d1 = gmtCalendar.getTime(); // 2008-10-01 0:00 GMT
        gmtCalendar.set(2008, Calendar.OCTOBER, 2, 0, 0, 0);
        Date d2 = gmtCalendar.getTime(); // 2008-10-02 0:00 GMT
        gmtCalendar.set(2008, Calendar.OCTOBER, 3, 0, 0, 0);
        Date d3 = gmtCalendar.getTime(); // 2008-10-03 0:00 GMT
        
        Convertor c = Task2Test.merge(
            limitTo(Task1Test.createCZKtoUSD(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        Date date;
        BigDecimal amount;

        // convert $5 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try
        {
            c.convert(Task1Test.USD, Task1Test.CZK, new BigDecimal("5.00"));
            fail("test A");
        }
        catch(final InvalidConversionException ex)
        {
        }

        // convert $8 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try
        {
            c.convert(Task1Test.USD, Task1Test.CZK, new BigDecimal("8.00"));
            fail("test B");
        }
        catch(final InvalidConversionException ex)
        {
        }

        // convert 1003CZK to USD using c:
        // cannot convert as no rate is applicable to current date
        try
        {
            c.convert(Task1Test.CZK, Task1Test.USD, new BigDecimal("1003.00"));
            fail("test C");
        }
        catch(final InvalidConversionException ex)
        {
        }

        // convert 16CZK using c:
        // cannot convert as no rate is applicable to current date
        try
        {
            c.convert(Task1Test.CZK, Task1Test.USD, new BigDecimal("16.00"));
            fail("test D");
        }
        catch(final InvalidConversionException ex)
        {
        }

        // convert 500SKK to CZK using c:
        // cannot convert as no rate is applicable to current date
        try
        {
            c.convert(Task1Test.SKK, Task1Test.CZK, new BigDecimal("500.00"));
            fail("test C");
        }
        catch(final InvalidConversionException ex)
        {
        }

        // convert $5 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 85 CZK");
        gmtCalendar.set(2008, Calendar.OCTOBER, 1, 6, 0, 0);
        date = gmtCalendar.getTime();
        amount = ((TimedConvertor)c).convert(Task1Test.USD, Task1Test.CZK, new BigDecimal("5.00"), date);
        assertEquals(new BigDecimal("85.00"), amount);

        // convert $8 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 136 CZK");
        gmtCalendar.set(2008, Calendar.OCTOBER, 1, 6, 0, 0);
        date = gmtCalendar.getTime();
        amount = ((TimedConvertor)c).convert(Task1Test.USD, Task1Test.CZK, new BigDecimal("8.00"), date);
        assertEquals(new BigDecimal("136.00"), amount);

        // convert 1003CZK to USD using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 59 USD");
        gmtCalendar.set(2008, Calendar.OCTOBER, 1, 6, 0, 0);
        date = gmtCalendar.getTime();
        amount = ((TimedConvertor)c).convert(Task1Test.CZK, Task1Test.USD, new BigDecimal("1003.00"), date);
        assertEquals(new BigDecimal("59.00"), amount);

        // convert 16CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 20 SKK");
        gmtCalendar.set(2008, Calendar.OCTOBER, 2, 9, 0, 0);
        date = gmtCalendar.getTime();
        amount = ((TimedConvertor)c).convert(Task1Test.CZK, Task1Test.SKK, new BigDecimal("16.00"), date);
        assertEquals(new BigDecimal("20.00"), amount);

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 400 CZK");
        gmtCalendar.set(2008, Calendar.OCTOBER, 2, 9, 0, 0);
        date = gmtCalendar.getTime();
        amount = ((TimedConvertor)c).convert(Task1Test.SKK, Task1Test.CZK, new BigDecimal("500.00"), date);
        assertEquals(new BigDecimal("400.00"), amount);

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        // cannot convert as no rate is applicable to current date
        try
        {
            gmtCalendar.set(2008, Calendar.OCTOBER, 1, 6, 0, 0);
            date = gmtCalendar.getTime();
            ((TimedConvertor)c).convert(Task1Test.SKK, Task1Test.CZK, new BigDecimal("500.00"), date);
            fail("test D");
        }
        catch(final InvalidConversionException ex)
        {
        }
    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 90 CZK.
     *
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor createSKKtoCZK2()
    {
        final ExchangeRate rate;
        final Convertor    convertor;
        
        rate = ExchangeRate.getExchangeRate(Task1Test.SKK,
                                            Task1Test.CZK,
                                            new BigDecimal("100.00"),
                                            new BigDecimal("90.00"));
        convertor = ConvertorFactory.getConvertor(rate);

        return (convertor);
    }

    public void testDateConvetorWithTwoDifferentRates() throws Exception {
        gmtCalendar.set(2008, Calendar.OCTOBER, 1, 0, 0, 0);
        Date d1 = gmtCalendar.getTime(); // 2008-10-01 0:00 GMT
        gmtCalendar.set(2008, Calendar.OCTOBER, 2, 0, 0, 0);
        Date d2 = gmtCalendar.getTime(); // 2008-10-02 0:00 GMT
        gmtCalendar.set(2008, Calendar.OCTOBER, 3, 0, 0, 0);
        Date d3 = gmtCalendar.getTime(); // 2008-10-03 0:00 GMT

        Convertor c = Task2Test.merge(
            limitTo(createSKKtoCZK2(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        Date date;
        BigDecimal amount;
        
        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 400 CZK");
        gmtCalendar.set(2008, Calendar.OCTOBER, 2, 6, 0, 0);
        date = gmtCalendar.getTime();
        amount = ((TimedConvertor)c).convert(Task1Test.SKK, Task1Test.CZK, new BigDecimal("500.00"), date);
        assertEquals(new BigDecimal("400.00"), amount);

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 450 CZK");
        gmtCalendar.set(2008, Calendar.OCTOBER, 1, 6, 0, 0);
        date = gmtCalendar.getTime();
        amount = ((TimedConvertor)c).convert(Task1Test.SKK, Task1Test.CZK, new BigDecimal("500.00"), date);
        assertEquals(new BigDecimal("450.00"), amount);
    }

}
