package org.apidesign.apifest08.test;

import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.Date;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.UnsupportedConversionException;

import static org.apidesign.apifest08.test.Currencies.*;

/** The exchange rates are not always the same. They are changing. However
 * as in order to predict the future, one needs to understand own past. That is
 * why it is important to know the exchange rate as it was at any time during
 * the past.
 * <p>
 * Today's quest is to enhance the convertor API to deal with dates.
 * One shall be able to convert a currency at any date. Each currencies rate shall
 * be associated with a range between two Date objects. In order
 * to keep compatibility with old API that knew nothing about dates, the
 * rates associated then are applicable "for eternity". Any use of existing
 * convert methods that do not accept a Date argument, uses the current
 * System.currentTimeMillis() as default date.
 */
public class Task4Test extends TestCase {
    public Task4Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //

    /** Takes a convertor with any rates associated and creates new convertor
     * that returns the same values as the old one for time between from to till.
     * Otherwise it returns no results. This is just a helper method that
     * shall call some real one in the API.
     * 
     * @param old existing convertor
     * @param from initial date (inclusive)
     * @param till final date (exclusive)
     * @return new convertor
     */
    public static Convertor limitTo(Convertor old, Date from, Date till) {
        return new Convertor(old, from, till);
    }


    public void testCompositionOfLimitedConvertors() throws Exception {
        SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd HH:mm zzzz");
        
        Date d1 = df.parse("2008-10-01 0:00 GMT"); 
        Date d2 = df.parse("2008-10-02 0:00 GMT"); 
        Date d3 = df.parse("2008-10-03 0:00 GMT");
        
        Convertor c = Task2Test.merge(
            limitTo(Task1Test.createCZKtoUSD(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert $5 to CZK using c:
        try {
        	c.convert(new BigDecimal(5), USD , CZK);
        	fail("cannot convert as no rate is applicable to current date");
        } catch(UnsupportedConversionException e) {
        	//expected
        }

        // convert $8 to CZK using c:
        try {
        	c.convert(new BigDecimal(8), USD , CZK);
        	fail("cannot convert as no rate is applicable to current date");
        } catch(UnsupportedConversionException e) {
        	//expected
        }

        // convert 1003CZK to USD using c:
        try {
        	c.convert(new BigDecimal(1003), CZK, USD);
        	fail("cannot convert as no rate is applicable to current date");
        } catch(UnsupportedConversionException e) {
        	//expected
        }

        // convert 16CZK using c:
        try {
        	c.convert(new BigDecimal(16), CZK, USD);
        	fail("cannot convert as no rate is applicable to current date");
        } catch(UnsupportedConversionException e) {
        	//expected
        }

        // convert 500SKK to CZK using c:
        try {
        	c.convert(new BigDecimal(500), SKK, CZK);
        	fail("cannot convert as no rate is applicable to current date");
        } catch(UnsupportedConversionException e) {
        	//expected
        }

        // convert $5 to CZK using c at 2008-10-01 6:00 GMT:         
        assertEquals("Result is 85 CZK", 85, c.convert(new BigDecimal(5), USD, CZK, df.parse("2008-10-01 6:00 GMT")).getValue().intValue());

        // convert $8 to CZK using c at 2008-10-01 6:00 GMT:
        assertEquals("Result is 136 CZK", 136, c.convert(new BigDecimal(8), USD, CZK, df.parse("2008-10-01 6:00 GMT")).getValue().intValue());
     
        // convert 1003CZK to USD using c at 2008-10-01 6:00 GMT:
        assertEquals("Result is 59 USD", 59, c.convert(new BigDecimal(1003), CZK, USD, df.parse("2008-10-01 6:00 GMT")).getValue().intValue());

        // convert 16CZK using c at 2008-10-02 9:00 GMT:
        assertEquals("Result is 20 SKK", 20, c.convert(new BigDecimal(16), CZK, SKK, df.parse("2008-10-02 9:00 GMT")).getValue().intValue());

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        assertEquals("Result is 400 CZK", 400, c.convert(new BigDecimal(500), SKK, CZK, df.parse("2008-10-02 9:00 GMT")).getValue().intValue());

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        try {
        	c.convert(new BigDecimal(500), SKK, CZK, df.parse("2008-10-01 6:00 GMT"));
        	fail("cannot convert as no rate is applicable to current date");
        } catch(UnsupportedConversionException e) {
        	//expected
        }
    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 90 CZK.
     *
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor createSKKtoCZK2() {
    	return new Convertor(new BigDecimal("0.9"), SKK, CZK);
    }

    public void testDateConvetorWithTwoDifferentRates() throws Exception {
        SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd HH:mm zzzz");
        
        Date d1 = df.parse("2008-10-01 0:00 GMT"); 
        Date d2 = df.parse("2008-10-02 0:00 GMT"); 
        Date d3 = df.parse("2008-10-03 0:00 GMT");

        Convertor c = Task2Test.merge(
            limitTo(createSKKtoCZK2(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        assertEquals("Result is 400 CZK", 400, c.convert(new BigDecimal(500), SKK, CZK, df.parse("2008-10-02 9:00 GMT")).getValue().intValue());

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        assertEquals("Result is 450 CZK", 450, c.convert(new BigDecimal(500), SKK, CZK, df.parse("2008-10-01 6:00 GMT")).getValue().intValue());
    }

}
