package org.apidesign.apifest08.test;

import java.math.BigDecimal;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.ConversionResult;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.ConvertorCurrency;
import org.apidesign.apifest08.currency.ExchangeRateProvider;
import org.apidesign.apifest08.currency.ConversionNotSupportedException;

/** Finish the Convertor API, and then write bodies of methods inside
 * of this class to match the given tasks. To fullfil your task, use the
 * API define in the <code>org.apidesign.apifest08.currency</code> package.
 * Do not you reflection, or other hacks as your code
 * shall run without any runtime permissions.
 */
public class Task1Test extends TestCase {
    public Task1Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    /** Create convertor that understands two currencies, CZK and
     *  USD. Make 1 USD == 17 CZK.
     *
     * Creation of the convertor shall not require subclassing of any class
     * or interface on the client side.
     *
     * @return prepared convertor ready for converting USD to CZK and CZK to USD
     */
    public static Convertor createCZKtoUSD() {
        ConvertorCurrency fromCurrency = ConvertorCurrency.getInstance("CZK");
        ConvertorCurrency toCurrency = ConvertorCurrency.getInstance("USD");
        ExchangeRateProvider exchangeRateProvider = new ExchangeRateProvider(new BigDecimal(17), fromCurrency, new BigDecimal(1), toCurrency);
        
        return Convertor.createConvertor(exchangeRateProvider);
    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 80 CZK.
     *
     * Creation of the convertor shall not require subclassing of any class
     * or interface on the client side.
     * 
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor createSKKtoCZK() {
        ConvertorCurrency fromCurrency = ConvertorCurrency.getInstance("SKK");
        ConvertorCurrency toCurrency = ConvertorCurrency.getInstance("CZK");
        ExchangeRateProvider exchangeRateProvider = new ExchangeRateProvider(new BigDecimal(100), fromCurrency, new BigDecimal(80), toCurrency);
        
        return Convertor.createConvertor(exchangeRateProvider);
    }
    
    
    public static Convertor createCZKtoYEN() {
        ConvertorCurrency fromCurrency = ConvertorCurrency.getInstance("CZK");
        ConvertorCurrency toCurrency = ConvertorCurrency.getInstance("JPY");
        ExchangeRateProvider exchangeRateProvider = new ExchangeRateProvider(new BigDecimal(1), fromCurrency, new BigDecimal(1), toCurrency);
        
        return Convertor.createConvertor(exchangeRateProvider);
    }
    
    /** Use the convertor from <code>createCZKtoUSD</code> method and do few conversions
     * with it.
     */
    public void testCurrencyCZKUSD() throws Exception {
        Convertor convertCzkUsd = createCZKtoUSD();

        {
            // convert $1 to CZK using c:
            ConversionResult result = convertCzkUsd.convertBack(new BigDecimal(1));
            assertEquals("Result is 17 CZK", new BigDecimal("17.00"), result.getConverted());
            assertEquals("No Remainer", new BigDecimal("00.00").setScale(2), result.getRemainder());
        }

        {
            // convert 17CKZ to $ using c:
            ConversionResult result = convertCzkUsd.convert(new BigDecimal(17));
            assertEquals("Result is 1 $", new BigDecimal("1.00"), result.getConverted());
            assertEquals("No Remainer", new BigDecimal("00.00").setScale(2), result.getRemainder());
        }

        {
            // convert $5 to CZK using c:
            ConversionResult result = convertCzkUsd.convertBack(new BigDecimal(5));
            assertEquals("Result is 85 CZK", new BigDecimal("85.00"), result.getConverted());
            assertEquals("No Remainer", BigDecimal.ZERO.setScale(2), result.getRemainder());
        }
        
        {
            // convert $8 to CZK
            ConversionResult result = convertCzkUsd.convertBack(new BigDecimal(8));
            assertEquals("Result is 136 CZK", new BigDecimal("136.00"), result.getConverted());
            assertEquals("No Remainer", BigDecimal.ZERO.setScale(2), result.getRemainder());
        }

        {
            // convert 1003CZK to USD
            ConversionResult result = convertCzkUsd.convert(new BigDecimal(1003));
            assertEquals("Result is 59 USD", new BigDecimal("59.00"), result.getConverted());
            assertEquals("No Remainer", BigDecimal.ZERO.setScale(2), result.getRemainder());
        }
    }

    /** Use the convertor from <code>createSKKtoCZK</code> method and do few conversions
     * with it.
     */
    public void testCurrencySKKCZK() throws Exception {
        Convertor convertSkkCzk = createSKKtoCZK();
        {
            // convert 100SKK using c:
            ConversionResult result = convertSkkCzk.convert(new BigDecimal(100));
            assertEquals("Result is 80 CZK", new BigDecimal("80.00"), result.getConverted());
        }
        {
            // convert 80CZK using c:
            ConversionResult result = convertSkkCzk.convertBack(new BigDecimal(80));
            assertEquals("Result is 100 SKK", new BigDecimal("100.00"), result.getConverted());
        }
        
        {
            // convert 16CZK using c:
            ConversionResult result = convertSkkCzk.convertBack(new BigDecimal(16));
            assertEquals("Result is 20 SKK", new BigDecimal("20.00"), result.getConverted());
        }

        {
            // convert 500SKK to CZK
            ConversionResult result = convertSkkCzk.convert(new BigDecimal(500));
            assertEquals("Result is 400 CZK", new BigDecimal("400.00"), result.getConverted());
            assertEquals("No Remainer", BigDecimal.ZERO.setScale(2), result.getRemainder());            
        }
        
        {
            // convert 501SKK to CZK
            ConversionResult result = convertSkkCzk.convert(new BigDecimal(501));
            assertEquals("Result is 400 CZK", new BigDecimal("400.80"), result.getConverted());
            assertEquals("No Remainer", BigDecimal.ZERO.setScale(2), result.getRemainder());
            
        }
    }
    
    /**
     * Convert SKK to CZK. Convertor can't convert whole amout (can't convert one SKK cent to CZK). Remaining
     * amount is stored in remainder result.
     * 
     * Test is currently failing, because implementation can't handle this case.
     */
//    public void testConvertSmallUnits_failing() {
//        Convertor convertSkkCzk = createSKKtoCZK();
//        {
//            // convert 501SKK to CZK
//            ConversionResult result = convertSkkCzk.convert(new BigDecimal("501.01"));
//            assertEquals("Result is 400 CZK", new BigDecimal("400.80"), result.getConverted());
//            assertEquals("No Remainer", new BigDecimal("0.01"), result.getRemainder());
//            
//        }
//        
//    }
    
    /**
     * Test converting from CZK to JPY. Remained has scale of CZK.
     * 
     * This test is currently failing, because converter implementation currently can't handle conversion from "cent" to "no-cent" currency.
     */
//    public void testConvertCzkToJpy_failing() {
//        Convertor convertSkkCzk = createCZKtoYEN();
//        {
//            // convert 501SKK to CZK
//            ConversionResult result = convertSkkCzk.convert(new BigDecimal("120.00"));
//            assertEquals("Result is 120 YEN", new BigDecimal("120"), result.getConverted());
//            assertEquals("No Remainer", new BigDecimal("0.00"), result.getRemainder());
//            
//        }
//    }
    
    /**
     * Test converting from JPY to CZK. Remained has scale of JPY.
     * 
     * This test is currently failing, because converter implementation currently can't handle conversion from "cent" to "no-cent" currency.
     */
//    public void testConvertJpyToCzk_failing() {
//        Convertor convertSkkCzk = createCZKtoYEN();
//        {
//            // convert 501SKK to CZK
//            ConversionResult result = convertSkkCzk.convert(new BigDecimal("120.00"));
//            assertEquals("Result is 120 YEN", new BigDecimal("120"), result.getConverted());
//            assertEquals("No Remainer", new BigDecimal("0"), result.getRemainder());
//            
//        }
//    }
    
    public void testCannotConvertToSKKwithCZKUSDConvertor() throws Exception {
            Convertor c = createCZKtoUSD();
            // convert $5 to SKK, the API shall say this is not possible
            try {
                c.convert(ConvertorCurrency.getInstance("USD"), ConvertorCurrency.getInstance("SKK"), new BigDecimal(5));
                fail();
            } catch (ConversionNotSupportedException e) {
                //expected error;
                assertEquals("Exception From USD", "USD",e.getFromCurrecyCode());
                assertEquals("Exception To SKK", "SKK",e.getToCurrecyCode());
            }
            // convert 500 SKK to CZK, the API shall say this is not possible
            try {
                c.convert(ConvertorCurrency.getInstance("SKK"), ConvertorCurrency.getInstance("CZK"), new BigDecimal(500));
                fail();
            } catch (ConversionNotSupportedException e) {
                assertEquals("Exception From USD", "SKK",e.getFromCurrecyCode());
                assertEquals("Exception To SKK", "CZK",e.getToCurrecyCode());            }
    }
    
 
}

