package org.apidesign.apifest08.test;

import java.math.BigDecimal;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.ConvertorCurrency;
import org.apidesign.apifest08.currency.ExchangeRateProvider;
import org.apidesign.apifest08.currency.IExchangeRateEngine;

/** The exchange rates are not always the same. They are changing. Day by day,
 * hour by hour, minute by minute. For every bank it is important to always
 * have the actual exchange rate available in the system. That is why let's
 * create a pluggable convertor that will always have up to date value of its
 * exchange rate.
 * <p>
 * The quest for today is to allow 3rd party developer to write a convertor
 * that adjusts its exchange rate everytime it is queried. This convertor is
 * written by independent vendor, the vendor knows only your Convertor API,
 * he does not know how the whole system looks and how the convertor is supposed
 * to be used.
 */
public class Task3Test extends TestCase {
    
    private static ConvertorCurrency currencyCZK = ConvertorCurrency.getInstance("CZK");
    private static ConvertorCurrency currencySKK = ConvertorCurrency.getInstance("SKK");
    private static ConvertorCurrency currencyUSD = ConvertorCurrency.getInstance("USD");
    
    public Task3Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //


    /** Without knowing anything about the surrounding system, write an
     * implementation of convertor that will return different rates everytime
     * it is queried. Convert USD to CZK and vice versa. Start with the rate of
     * 1USD = 16CZK and adjust it in favor of CZK by 0.01 CZK with every query.
     * As soon as you reach 1USD = 15CZK adjust it by 0.01 CZK in favor of USD
     * until you reach 1USD = 16CZK
     *
     * @return new instance of "online" USD and CZK convertor starting with rate 1USD = 16CZK
     */
    public static Convertor createOnlineCZKUSDConvertor() {
        // initial rate: 1USD = 16CZK
        // 2nd query 1USD = 15.99CZK
        // 3rd query 1USD = 15.98CZK
        // until 1USD = 15.00CZK
        // then 1USD = 15.01CZK
        // then 1USD = 15.02CZK
        // and so on and on up to 1USD = 16CZK
        // and then another round to 15, etc.
        IExchangeRateEngine engine = BouncingExchangeRateEngine.create(
                currencyUSD,  new BigDecimal(1),
                currencyCZK, 
                new BigDecimal(16),new BigDecimal(15),
                new BigDecimal("-0.01"));
        ExchangeRateProvider exRateProvider = ExchangeRateProvider.createExchangeRateProvider(engine);
        Convertor convertor = Convertor.createConvertor(exRateProvider);
        return convertor;
    }

    public void testFewQueriesForOnlineConvertor() {
        Convertor c = createOnlineCZKUSDConvertor();
        doFewQueriesForOnlineConvertor(c);
    }

    static void doFewQueriesForOnlineConvertor(Convertor c) {
        // convert $5 to CZK using c:
        assertEquals("Result is 80 CZK", new BigDecimal("80.00"), c.convertWithReversibleRates(currencyUSD, currencyCZK, new BigDecimal(5)).getConverted());

        // convert $8 to CZK using c:
        assertEquals("Result is 127.92 CZK", new BigDecimal("127.92"), c.convertWithReversibleRates(currencyUSD, currencyCZK, new BigDecimal(8)).getConverted());

        // convert $1 to CZK using c:
        assertEquals("Result is 15.98 CZK", new BigDecimal("15.98"), c.convertWithReversibleRates(currencyUSD, currencyCZK, new BigDecimal(1)).getConverted());

        // convert 15.97CZK to USD using c:
        assertEquals("Result is 15.98 CZK", new BigDecimal("1.00"), c.convertWithReversibleRates(currencyCZK, currencyUSD , new BigDecimal("15.97")).getConverted());
    }

    /** Join the convertors and show they behave sane.
     */
    public void testOnlineConvertorComposition() throws Exception {
        Convertor c = Task2Test.merge(
            createOnlineCZKUSDConvertor(),
            Task1Test.createSKKtoCZK()
        );

        // convert 16CZK to SKK using c:
        assertEquals("Result is 20 CZK", new BigDecimal("20.00"), c.convertWithReversibleRates(currencyCZK, currencySKK , new BigDecimal("16")).getConverted());

        // convert 500SKK to CZK using c:
        assertEquals("Result is 400 CZK", new BigDecimal("400.00"), c.convertWithReversibleRates(currencySKK, currencyCZK , new BigDecimal("500")).getConverted());

        doFewQueriesForOnlineConvertor(c);
    }
    
//    public void testBouncing() {
//        Convertor c=createOnlineCZKUSDConvertor();
//        
//        IExchangeRateEngine engine = BouncingExchangeRateEngine.create(
//                currencyUSD,  new BigDecimal(1),
//                currencyCZK, 
//                new BigDecimal("16.00"),new BigDecimal("15.00"),
//                new BigDecimal("-0.01"));        
//        
//        for (int i=0;i<300;i++) {
//            System.out.println(engine.getExchangeRate(currencyUSD, currencyCZK).getToValue());
//        }
//    }
}
